import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

public class Algorithms_LotSizing {
	static int inventory;
	static String[] statuses; //"open", "served"
	
	static double solveInstance(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_LotSizing instance, int lookahead){
		double costs = 0;
		inventory = 0;
		int T = instance.T;
		statuses = new String[T];
		for(int t = 0; t < T; t++)
			statuses[t] = "open";
		int t = 0;
		do{
			int numberOfPeriods = coverDemands(algorithm, gradualLossFactor, maxGradualHorizon, instance, lookahead, t);
			if(numberOfPeriods > lookahead){
				instance.createForecasts(gradualLossFactor, t, lookahead);
				boolean shortageFound = false;
				for(int t_later = t + lookahead; t_later < t + numberOfPeriods && !shortageFound; t_later++)
					if(instance.demands[t_later] > instance.demands_forecasted[t_later]){
						numberOfPeriods = t_later - t;
						shortageFound = true;
					}
			}
			costs += bookkeeping(instance, lookahead, t, numberOfPeriods);
			t += numberOfPeriods;
		}
		while(t < instance.T);
		return costs;
	}
	
	static double bookkeeping(Instance_LotSizing instance, int lookahead, int t_current, int numberOfPeriods){
		double additionalCosts = instance.A;
		for(int t = t_current; t < t_current + numberOfPeriods; t++)
			if(t < t_current + lookahead)
				inventory += instance.demands[t];
			else
				inventory += instance.demands_forecasted[t];
		for(int t = t_current; t < t_current + numberOfPeriods; t++){
			statuses[t] = "served";
			inventory -= instance.demands[t];
			additionalCosts += instance.h * inventory;
		}
		return additionalCosts;
	}
	
	static int coverDemands(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_LotSizing instance, int lookahead, int t_current){
		int numberOfPeriods = -1;
		if(algorithm.equals("SM"))
			numberOfPeriods = getNumberOfPeriods_SM(instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIP"))
			numberOfPeriods = getNumberOfPeriods_SnapshotMIP(false, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIPForecasts"))
			numberOfPeriods = getMostFrequentNumberOfPeriods(gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		return numberOfPeriods;
	}
	
	static int getMostFrequentNumberOfPeriods(double gradualLossFactor, int maxGradualHorizon, Instance_LotSizing instance, int lookahead, int t_current){
		int numberOfShootings = 25;
		int[] numbersOfPeriods = new int[numberOfShootings];
		for(int it = 0; it < numberOfShootings; it++)
			numbersOfPeriods[it] = getNumberOfPeriods_SnapshotMIP(true, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		return (new Distribution(numbersOfPeriods)).getValueWithHighestProbability();
	}
	
	static int getNumberOfPeriods_SM(Instance_LotSizing instance, int lookahead, int t_current){
		int n = Math.min(lookahead, instance.T - t_current);
		double[] x = new double[n];
		int t = 0;
		double lastPeriodCovered = t;
		while(lastPeriodCovered < n - 1){
			x[t] += instance.demands[t_current + t];
			lastPeriodCovered = t;
			boolean increasingQuantityWorthwhile = true;
			for(int s = t + 1; s < t + n && s < n && increasingQuantityWorthwhile; s++){
				double sum_new = 0;
				for(int j = t + 1; j <= s; j++)
					sum_new += (j - t) * instance.demands[t_current + j];
				double C_new = (1.0 / (s - t + 1)) * (instance.A + instance.h * sum_new);
				double sum_old = 0;
				for(int j = t + 1; j <= s-1; j++)
					sum_old += (j - t) * instance.demands[t_current + j];
				double C_old = (1.0 / (s - 1 - t + 1)) * (instance.A + instance.h * sum_old);
				if(C_new > C_old){
					increasingQuantityWorthwhile = false;
					t = s;
				}
				else{
					x[t] += instance.demands[t_current + s];
					lastPeriodCovered = s;
				}
				if(s == t + n - 1 && increasingQuantityWorthwhile)
					t = s;
			}
		}
		for(t = 1; t < n; t++)
			if(x[t] > 0)
				return t;
		return n;
	}	
	
	static int getNumberOfPeriods_SnapshotMIP(boolean forecasts, double gradualLossFactor, int maxGradualHorizon, Instance_LotSizing instance, int lookahead, int t_current){
		int n_lookahead = Math.min(lookahead, instance.T - t_current);
		int n_forecasts;
		if(!forecasts)
			n_forecasts = 0;
		else{
			n_forecasts = Math.min(instance.T - t_current - n_lookahead, maxGradualHorizon);
			instance.createForecasts(gradualLossFactor, t_current, lookahead);
		}
		int n = n_lookahead + n_forecasts;

		double[] d = new double[n];
		for(int i = 0; i < n; i++)
			if(i < n_lookahead)
				d[i] = instance.demands[t_current + i];
			else
				d[i] = instance.demands_forecasted[t_current + i];

		int numberOfPeriods = -1;
		try{
			IloCplex cplex = new IloCplex();
			IloNumVar[] x = cplex.numVarArray(n, 0, 9999);
			IloNumVar[] y = cplex.boolVarArray(n);
			IloNumVar[] I = cplex.numVarArray(n, 0, 9999);
			
			IloLinearNumExpr objective = cplex.linearNumExpr();
			for(int i = 0; i < n; i++){
				objective.addTerm(instance.A, y[i]);
				objective.addTerm(instance.h, I[i]);
			}
	        cplex.addMinimize(objective);
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
       			expr.addTerm(1, I[i]);
       			if(i > 0)
       				expr.addTerm(-1, I[i-1]);
	        	expr.addTerm(-1, x[i]);
	        	if(i > 0)
	        		cplex.addEq(expr, -d[i]);
	        	else
	        		cplex.addEq(expr, inventory-d[i]);
	        }
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
       			expr.addTerm(1, x[i]);
	        	expr.addTerm(-99999, y[i]);
        		cplex.addLe(expr, 0);
	        }
     	
	        cplex.setOut(null);
	        cplex.setParam(IloCplex.DoubleParam.EpGap, 0.05);
	        cplex.setParam(IloCplex.DoubleParam.TiLim, 60);
	        cplex.solve();
	        boolean firstFound = false;
	        for(int i = 1; i < n && !firstFound; i++)
	        	if(Math.abs(cplex.getValue(y[i]) - 1) < 0.0001){
	        		numberOfPeriods = i;
	        		firstFound = true;
	        	}
	        if(numberOfPeriods == -1)
	        	numberOfPeriods = n;
	        cplex.end();
		}
	    catch (IloException ex){
	    	System.out.println("Concert Error: " + ex);
	    }
		return numberOfPeriods;
	}
}