import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

class Edge {
	int start, end;
	public Edge(int start, int end){
		this.start = start;
		this.end = end;
	}
}

public class Algorithms_Scheduling {
	static int inventory;
	static String[] statuses; //"open", "accepted", "rejected"
	
	static double solveInstance(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_Scheduling instance, int lookahead){
		double costs = 0;
		inventory = 0;
		int T = instance.T;
		statuses = new String[T];
		for(int t = 0; t < T; t++)
			statuses[t] = "open";
		boolean decision;
		for(int t = 0; t < T; t++){
			if(statuses[t].equals("open")){
				decision = decideOnNextJob(algorithm, gradualLossFactor, maxGradualHorizon, instance, lookahead, t);
				costs += bookkeeping(instance, lookahead, t, decision);
			}
		}
		return costs;
	}
	
	static double bookkeeping(Instance_Scheduling instance, int lookahead, int t_current, boolean decision){
		double additionalCosts = decision ? instance.processingTimes[t_current] : 0;
		statuses[t_current] = decision ? "accepted" : "rejected";
		if(statuses[t_current].equals("accepted"))
			for(int t = t_current + 1; t < instance.T; t++)
				if(instance.releaseDates[t_current] + instance.processingTimes[t_current] > instance.releaseDates[t])
					statuses[t] = "rejected";
		return additionalCosts;
	}
	
	static boolean decideOnNextJob(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_Scheduling instance, int lookahead, int t_current){
		boolean decision = false;
		if(algorithm.equals("Heuristic"))
			decision = getAcceptanceDecision_Heuristic(instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIP"))
			decision = getAcceptanceDecision_SnapshotMIP(false, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIPForecasts"))
			decision = getMostFrequentDecision(gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		return decision;
	}
	
	static boolean getMostFrequentDecision(double gradualLossFactor, int maxGradualHorizon, Instance_Scheduling instance, int lookahead, int t_current){
		int numberOfShootings = 25;
		boolean[] decisions = new boolean[numberOfShootings];
		for(int it = 0; it < numberOfShootings; it++)
			decisions[it] = getAcceptanceDecision_SnapshotMIP(true, gradualLossFactor, maxGradualHorizon, instance, lookahead, t_current);
		int[] decisionsBinary = new int[numberOfShootings];
		for(int it = 0; it < numberOfShootings; it++)
			decisionsBinary[it] = decisions[it] ? 1 : 0;
		return (new Distribution(decisionsBinary)).getValueWithHighestProbability() == 1 ? true : false;
	}
	
	static boolean getAcceptanceDecision_Heuristic(Instance_Scheduling instance, int lookahead, int t_current){
		int n = Math.min(lookahead, instance.T - t_current);
		double[] r = new double[n];
		double[] p = new double[n];
		for(int i = 0; i < n; i++){
			r[i] = instance.releaseDates[t_current + i];
			p[i] = instance.processingTimes[t_current + i];
		}		
		boolean decision = false;
		ArrayList<Integer> overlappingIndizes = new ArrayList<Integer>();
		for(int i = 1; i < n; i++)
			if(r[i] < r[0] + p[0])
				overlappingIndizes.add(i);
		if(overlappingIndizes.size() == 0)
			decision = true;
		else{
			double avgReward = 0;
			for(int i = 0; i < overlappingIndizes.size(); i++)
				avgReward += p[overlappingIndizes.get(i)];
			avgReward = avgReward / overlappingIndizes.size();
			if(p[0] > avgReward)
				decision = true;
		}
		return decision;
	}
	
	/*
	static boolean getAcceptanceDecision_SnapshotMIP(boolean forecasts, double gradualLossFactor, Instance_Scheduling instance, int lookahead, int t_current){
		int n_lookahead = Math.min(lookahead, instance.T - t_current);
		int n_forecasts;
		if(!forecasts)
			n_forecasts = 0;
		else{
			n_forecasts = instance.T - t_current - n_lookahead;
			instance.createForecasts(gradualLossFactor, t_current, lookahead);
		}
		int n = n_lookahead + n_forecasts;

		double[] r = new double[n];
		double[] p = new double[n];
		for(int i = 0; i < n; i++){
			r[i] = instance.releaseDates[t_current + i];
			if(i < n_lookahead)
				p[i] = instance.processingTimes[t_current + i];
			else
				p[i] = instance.processingTimes_forecasted[t_current + i];
		}

		boolean decision = false;
		try{
			IloCplex cplex = new IloCplex();
			IloNumVar[][] x = new IloNumVar[n][];
			for(int i = 0; i < n; i++)
				x[i] = cplex.boolVarArray(1000);
			IloNumVar[] y = cplex.boolVarArray(n);
			
			IloLinearNumExpr objective = cplex.linearNumExpr();
			for(int i = 0; i < n; i++)
				objective.addTerm(p[i], y[i]);
	        cplex.addMaximize(objective);
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int t = 0; t < 1000; t++)
	        		expr.addTerm(t, x[i][t]);
	        	expr.addTerm(-r[i], y[i]);
        		cplex.addEq(expr, 0);
	        }
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int t = 0; t < 1000; t++)
	        		expr.addTerm(1, x[i][t]);
        		cplex.addLe(expr, 1);
	        }
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int t = 0; t < 1000; t++)
	        		expr.addTerm(1, x[i][t]);
	        	expr.addTerm(-1, y[i]);
        		cplex.addEq(expr, 0);
	        }
	        
	        for(int t = 0; t < 1000; t++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int i = 0; i < n; i++)
	        		for(int t_prime = 0; t_prime < 1000; t_prime++)
	        			if(t_prime < t && t_prime >= t - p[i])
	        				expr.addTerm(1, x[i][t_prime]);
	        	cplex.addLe(expr, 1);
	        }    
     	
	        cplex.setOut(null);
	        cplex.setParam(IloCplex.DoubleParam.EpGap, 0.05);
	        cplex.setParam(IloCplex.DoubleParam.TiLim, 60);
	        cplex.solve();
	        if(Math.abs(cplex.getValue(y[0]) - 1) < 0.001)
        		decision = true;
	        cplex.end();
		}
	    catch (IloException ex){
	    	System.out.println("Concert Error: " + ex);
	    }
		return decision;
	}
	*/
	
	static boolean getAcceptanceDecision_SnapshotMIP(boolean forecasts, double gradualLossFactor, int maxGradualHorizon, Instance_Scheduling instance, int lookahead, int t_current){
		int n_lookahead = Math.min(lookahead, instance.T - t_current);
		int n_forecasts;
		if(!forecasts)
			n_forecasts = 0;
		else{
			n_forecasts = Math.min(instance.T - t_current - n_lookahead, maxGradualHorizon);
			instance.createForecasts(gradualLossFactor, t_current, lookahead);
		}
		int n = n_lookahead + n_forecasts;

		double[] r = new double[n];
		double[] p = new double[n];
		for(int i = 0; i < n; i++){
			r[i] = instance.releaseDates[t_current + i];
			if(i < n_lookahead)
				p[i] = instance.processingTimes[t_current + i];
			else
				p[i] = instance.processingTimes_forecasted[t_current + i];
		}

		boolean decision = false;
		ArrayList<HashSet<Integer>> maximalCliques = computeMaximalCliquesOnInstance(r, p);
		int m = maximalCliques.size();
		int[][] A = new int[m][n];
		for(int i = 0; i < m; i++)
			for(int j = 0; j < n; j++)
				if(maximalCliques.get(i).contains(j))
					A[i][j] = 1;
		try{
			IloCplex cplex = new IloCplex();
			IloNumVar[] x = cplex.boolVarArray(n);			
			IloLinearNumExpr objective = cplex.linearNumExpr();
			for(int i = 0; i < n; i++)
				objective.addTerm(p[i], x[i]);
	        cplex.addMaximize(objective);	        
	        for(int i = 0; i < m; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int j = 0; j < n; j++)
	        		expr.addTerm(A[i][j], x[j]);
	        	cplex.addLe(expr, 1);
	        }	        
	        cplex.setOut(null);
	        cplex.setParam(IloCplex.DoubleParam.EpGap, 0.05);
	        cplex.setParam(IloCplex.DoubleParam.TiLim, 60);
	        cplex.solve();
	        if(Math.abs(cplex.getValue(x[0]) - 1) < 0.001)
	        	decision = true;
	        cplex.end();			
		}
	    catch (IloException ex){
	    	System.out.println("Concert Error: " + ex);
	    }
		return decision;
	}
	
	static ArrayList<HashSet<Integer>> computeMaximalCliquesOnInstance(double[] r, double[] p){
		ArrayList<Edge> edges = new ArrayList<Edge>();
		int n = r.length;
		boolean[][] adjacencyMatrix = new boolean[n][n];
		for(int i = 0; i < n; i++)
			for(int j = 0; j < n; j++){				
				if(i != j && r[i] <= r[j] && r[i]+p[i] <= r[j]+p[j] && r[j] <= r[i]+p[i])
					adjacencyMatrix[i][j] = true;
				if(i != j && r[i] >= r[j] && r[i]+p[i] >= r[j]+p[j] && r[j]+p[j] >= r[i])
					adjacencyMatrix[i][j] = true;
				if(i != j && r[i] <= r[j] && r[i]+p[i] >= r[j]+p[j])
					adjacencyMatrix[i][j] = true;
				if(i != j && r[i] >= r[j] && r[i]+p[i] <= r[j]+p[j])
					adjacencyMatrix[i][j] = true;
			}
		ArrayList<ArrayList<Integer>> adjacencyList = new ArrayList<ArrayList<Integer>>();
		for(int i = 0; i < n; i++)
			adjacencyList.add(new ArrayList<Integer>());		
		for(int i = 0; i < n; i++)
			for(int j = 0; j < n; j++)
				if(adjacencyMatrix[i][j] && i < j){
					edges.add(new Edge(i, j));
					adjacencyList.get(i).add(j);
					adjacencyList.get(j).add(i);
				}
		HashSet<Integer> R = new HashSet<Integer>();
		HashSet<Integer> X = new HashSet<Integer>();
		HashSet<Integer> P = new HashSet<Integer>();
		for(int i = 0; i < n; i++)
			P.add(i);
		ArrayList<HashSet<Integer>> maximalCliques = new ArrayList<HashSet<Integer>>();
		BronKerbosch(R, P, X, adjacencyList, maximalCliques);
		return maximalCliques;
	}
	
	static ArrayList<HashSet<Integer>> BronKerbosch(HashSet<Integer> R, HashSet<Integer> P, HashSet<Integer> X, ArrayList<ArrayList<Integer>> adjacencyList, ArrayList<HashSet<Integer>> maximalCliques){
		if(P.size() == 0 && X.size() == 0){
			HashSet<Integer> newClique = new HashSet<Integer>(R);
			maximalCliques.add(newClique);
		}
		else
			for(Iterator<Integer> iterator = P.iterator(); iterator.hasNext();){
				Integer v = iterator.next();
				HashSet<Integer> R_strich = new HashSet<Integer>(R);
				R_strich.add(v);
				HashSet<Integer> P_strich = new HashSet<Integer>(P);
				HashSet<Integer> X_strich = new HashSet<Integer>(X);
				HashSet<Integer> N_v = new HashSet<Integer>(adjacencyList.get(v));
				P_strich.retainAll(N_v);
				X_strich.retainAll(N_v);
				BronKerbosch(R_strich, P_strich, X_strich, adjacencyList, maximalCliques);
				iterator.remove();
				X.add(v);
			}
		return maximalCliques;
	}
}